from typing import Any
from pyquil.api._qpu import QPU as QPUV3
from ._qam import StatefulQAM


class QPU(QPUV3):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        StatefulQAM.wrap(self)



##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
from dataclasses import dataclass
import uuid
from collections import defaultdict
from typing import Dict, List, Optional, Union, cast

import numpy as np
from qcs_api_client.client import QCSClientConfiguration
from rpcq.messages import ParameterAref, ParameterSpec

from pyquil.api import QuantumExecutable, EncryptedProgram, EngagementManager

from pyquil._memory import Memory
from pyquil.api._qam import QAM, QAMExecutionResult
from pyquil.api._qpu_client import GetBuffersRequest, QPUClient, BufferResponse, RunProgramRequest
from pyquil.quilatom import (
    MemoryReference,
    BinaryExp,
    Function,
    Parameter,
    ExpressionDesignator,
)


def decode_buffer(buffer: BufferResponse) -> np.ndarray:
    """
    Translate a DataBuffer into a numpy array.

    :param buffer: Dictionary with 'data' byte array, 'dtype', and 'shape' fields
    :return: NumPy array of decoded data
    """
    buf = np.frombuffer(buffer.data, dtype=buffer.dtype)
    return buf.reshape(buffer.shape)  # type: ignore


def _extract_memory_regions(
    memory_descriptors: Dict[str, ParameterSpec],
    ro_sources: Dict[MemoryReference, str],
    buffers: Dict[str, np.ndarray],
) -> Dict[str, np.ndarray]:

    # hack to extract num_shots indirectly from the shape of the returned data
    first, *rest = buffers.values()
    num_shots = first.shape[0]

    def alloc(spec: ParameterSpec) -> np.ndarray:
        dtype = {
            "BIT": np.int64,
            "INTEGER": np.int64,
            "REAL": np.float64,
            "FLOAT": np.float64,
        }
        try:
            return np.ndarray((num_shots, spec.length), dtype=dtype[spec.type])
        except KeyError:
            raise ValueError(f"Unexpected memory type {spec.type}.")

    regions: Dict[str, np.ndarray] = {}

    for mref, key in ro_sources.items():
        # Translation sometimes introduces ro_sources that the user didn't ask for.
        # That's fine, we just ignore them.
        if mref.name not in memory_descriptors:
            continue
        elif mref.name not in regions:
            regions[mref.name] = alloc(memory_descriptors[mref.name])

        buf = buffers[key]
        if buf.ndim == 1:
            buf = buf.reshape((num_shots, 1))

        if np.iscomplexobj(buf):
            buf = np.column_stack((buf.real, buf.imag))
        _, width = buf.shape

        end = mref.offset + width
        region_width = memory_descriptors[mref.name].length
        if end > region_width:
            raise ValueError(
                f"Attempted to fill {mref.name}[{mref.offset}, {end})"
                f"but the declared region has width {region_width}."
            )

        regions[mref.name][:, mref.offset : end] = buf

    return regions


@dataclass
class QPUExecuteResponse:
    job_id: str
    _executable: EncryptedProgram


class QPU(QAM[QPUExecuteResponse]):
    def __init__(
        self,
        *,
        quantum_processor_id: str,
        priority: int = 1,
        timeout: float = 10.0,
        client_configuration: Optional[QCSClientConfiguration] = None,
        engagement_manager: Optional[EngagementManager] = None,
        endpoint_id: Optional[str] = None,
    ) -> None:
        """
        A connection to the QPU.

        :param quantum_processor_id: Processor to run against.
        :param priority: The priority with which to insert jobs into the QPU queue. Lower integers
            correspond to higher priority.
        :param timeout: Time limit for requests, in seconds.
        :param client_configuration: Optional client configuration. If none is provided, a default one will be loaded.
        :param endpoint_id: Optional endpoint ID to be used for engagement.
        :param engagement_manager: Optional engagement manager. If none is provided, a default one will be created.
        """
        super().__init__()

        self.priority = priority

        client_configuration = client_configuration or QCSClientConfiguration.load()
        engagement_manager = engagement_manager or EngagementManager(client_configuration=client_configuration)
        self._qpu_client = QPUClient(
            quantum_processor_id=quantum_processor_id,
            endpoint_id=endpoint_id,
            engagement_manager=engagement_manager,
            request_timeout=timeout,
        )
        self._last_results: Dict[str, np.ndarray] = {}
        self._memory_results: Dict[str, Optional[np.ndarray]] = defaultdict(lambda: None)

    @property
    def quantum_processor_id(self) -> str:
        """ID of quantum processor targeted."""
        return self._qpu_client.quantum_processor_id

    def execute(self, executable: QuantumExecutable) -> QPUExecuteResponse:
        """
        Enqueue a job for execution on the QPU. Returns a ``QPUExecuteResponse``, a
        job descriptor which should be passed directly to ``QPU.get_result`` to retrieve
        results.
        """
        executable = executable.copy()

        assert isinstance(
            executable, EncryptedProgram
        ), "QPU#execute requires an rpcq.EncryptedProgram. Create one with QuantumComputer#compile"

        assert (
            executable.ro_sources is not None
        ), "To run on a QPU, a program must include ``MEASURE``, ``CAPTURE``, and/or ``RAW-CAPTURE`` instructions"

        request = RunProgramRequest(
            id=str(uuid.uuid4()),
            priority=self.priority,
            program=executable.program,
            patch_values=self._build_patch_values(executable),
        )
        job_id = self._qpu_client.run_program(request).job_id
        return QPUExecuteResponse(_executable=executable, job_id=job_id)

    def get_result(self, execute_response: QPUExecuteResponse) -> QAMExecutionResult:
        """
        Retrieve results from execution on the QPU.
        """
        request = GetBuffersRequest(job_id=execute_response.job_id, wait=True)
        results = self._qpu_client.get_execution_results(request)

        ro_sources = execute_response._executable.ro_sources
        decoded_buffers = {k: decode_buffer(v) for k, v in results.buffers.items()}

        result_memory = {}
        if decoded_buffers is not None:
            extracted = _extract_memory_regions(
                execute_response._executable.memory_descriptors, ro_sources, decoded_buffers
            )
            for name, array in extracted.items():
                result_memory[name] = array
        elif not ro_sources:
            result_memory["ro"] = np.zeros((0, 0), dtype=np.int64)

        return QAMExecutionResult(
            executable=execute_response._executable,
            readout_data=result_memory,
            execution_duration_microseconds=results.execution_duration_microseconds,
        )

    @classmethod
    def _build_patch_values(cls, program: EncryptedProgram) -> Dict[str, List[Union[int, float]]]:
        """
        Construct the patch values from the program to be used in execution.
        """
        patch_values = {}

        # Now that we are about to run, we have to resolve any gate parameter arithmetic that was
        # saved in the executable's recalculation table, and add those values to the variables shim
        cls._update_memory_with_recalculation_table(program=program)

        # Initialize our patch table
        assert isinstance(program, EncryptedProgram)
        recalculation_table = program.recalculation_table
        memory_ref_names = list(set(mr.name for mr in recalculation_table.keys()))
        if memory_ref_names:
            assert len(memory_ref_names) == 1, (
                "We expected only one declared memory region for "
                "the gate parameter arithmetic replacement references."
            )
            memory_reference_name = memory_ref_names[0]
            patch_values[memory_reference_name] = [0.0] * len(recalculation_table)

        for name, spec in program.memory_descriptors.items():
            # NOTE: right now we fake reading out measurement values into classical memory
            # hence we omit them here from the patch table.
            if any(name == mref.name for mref in program.ro_sources):
                continue
            initial_value = 0.0 if spec.type == "REAL" else 0
            patch_values[name] = [initial_value] * spec.length

        # Fill in our patch table
        for k, v in program._memory.values.items():
            patch_values[k.name][k.index] = v

        return patch_values

    @classmethod
    def _update_memory_with_recalculation_table(cls, program: EncryptedProgram) -> None:
        """
        Update the program's memory with the final values to be patched into the gate parameters,
        according to the arithmetic expressions in the original program.

        For example::

            DECLARE theta REAL
            DECLARE beta REAL
            RZ(3 * theta) 0
            RZ(beta+theta) 0

        gets translated to::

            DECLARE theta REAL
            DECLARE __P REAL[2]
            RZ(__P[0]) 0
            RZ(__P[1]) 0

        and the recalculation table will contain::

            {
                ParameterAref('__P', 0): Mul(3.0, <MemoryReference theta[0]>),
                ParameterAref('__P', 1): Add(<MemoryReference beta[0]>, <MemoryReference theta[0]>)
            }

        Let's say we've made the following two function calls:

        .. code-block:: python

            compiled_program.write_memory(region_name='theta', value=0.5)
            compiled_program.write_memory(region_name='beta', value=0.1)

        After executing this function, our self.variables_shim in the above example would contain
        the following:

        .. code-block:: python

            {
                ParameterAref('theta', 0): 0.5,
                ParameterAref('beta', 0): 0.1,
                ParameterAref('__P', 0): 1.5,       # (3.0) * theta[0]
                ParameterAref('__P', 1): 0.6        # beta[0] + theta[0]
            }

        Once the _variables_shim is filled, execution continues as with regular binary patching.
        """
        assert isinstance(program, EncryptedProgram)
        for mref, expression in program.recalculation_table.items():
            # Replace the user-declared memory references with any values the user has written,
            # coerced to a float because that is how we declared it.
            program._memory.values[mref] = float(cls._resolve_memory_references(expression, memory=program._memory))

    @classmethod
    def _resolve_memory_references(cls, expression: ExpressionDesignator, memory: Memory) -> Union[float, int]:
        """
        Traverse the given Expression, and replace any Memory References with whatever values
        have been so far provided by the user for those memory spaces. Declared memory defaults
        to zero.

        :param expression: an Expression
        """
        if isinstance(expression, BinaryExp):
            left = cls._resolve_memory_references(expression.op1, memory=memory)
            right = cls._resolve_memory_references(expression.op2, memory=memory)
            return cast(Union[float, int], expression.fn(left, right))
        elif isinstance(expression, Function):
            return cast(
                Union[float, int],
                expression.fn(cls._resolve_memory_references(expression.expression, memory=memory)),
            )
        elif isinstance(expression, Parameter):
            raise ValueError(f"Unexpected Parameter in gate expression: {expression}")
        elif isinstance(expression, (float, int)):
            return expression
        elif isinstance(expression, MemoryReference):
            return memory.values.get(ParameterAref(name=expression.name, index=expression.offset), 0)
        else:
            raise ValueError(f"Unexpected expression in gate parameter: {expression}")



##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
from dataclasses import dataclass
from typing import Dict, Mapping, Optional, Sequence, Union, Tuple

import numpy as np
from qcs_api_client.client import QCSClientConfiguration

from pyquil._version import pyquil_version
from pyquil.api import QuantumExecutable
from pyquil.api._qam import QAM, QAMExecutionResult
from pyquil.api._qvm_client import (
    QVMClient,
    RunProgramRequest,
)
from pyquil.noise import NoiseModel, apply_noise_model
from pyquil.quil import Program, get_classical_addresses_from_program


class QVMVersionMismatch(Exception):
    pass


class QVMNotRunning(Exception):
    pass


def check_qvm_version(version: str) -> None:
    """
    Verify that there is no mismatch between pyquil and QVM versions.

    :param version: The version of the QVM
    """
    major, minor, patch = map(int, version.split("."))
    if major == 1 and minor < 8:
        raise QVMVersionMismatch(
            "Must use QVM >= 1.8.0 with pyquil >= 2.8.0, but you " f"have QVM {version} and pyquil {pyquil_version}"
        )


@dataclass
class QVMExecuteResponse:
    executable: Program
    memory: Mapping[str, np.ndarray]


class QVM(QAM[QVMExecuteResponse]):
    def __init__(
        self,
        noise_model: Optional[NoiseModel] = None,
        gate_noise: Optional[Tuple[float, float, float]] = None,
        measurement_noise: Optional[Tuple[float, float, float]] = None,
        random_seed: Optional[int] = None,
        timeout: float = 10.0,
        client_configuration: Optional[QCSClientConfiguration] = None,
    ) -> None:
        """
        A virtual machine that classically emulates the execution of Quil programs.

        :param noise_model: A noise model that describes noise to apply when emulating a program's
            execution.
        :param gate_noise: A tuple of three numbers [Px, Py, Pz] indicating the probability of an X,
           Y, or Z gate getting applied to each qubit after a gate application or reset. The
           default value of None indicates no noise.
        :param measurement_noise: A tuple of three numbers [Px, Py, Pz] indicating the probability
            of an X, Y, or Z gate getting applied before a measurement. The default value of
            None indicates no noise.
        :param random_seed: A seed for the QVM's random number generators. Either None (for an
            automatically generated seed) or a non-negative integer.
        :param timeout: Time limit for requests, in seconds.
        :param client_configuration: Optional client configuration. If none is provided, a default one will be loaded.
        """
        super().__init__()

        if (noise_model is not None) and (gate_noise is not None or measurement_noise is not None):
            raise ValueError(
                """
You have attempted to supply the QVM with both a Kraus noise model
(by supplying a `noise_model` argument), as well as either `gate_noise`
or `measurement_noise`. At this time, only one may be supplied.

To read more about supplying noise to the QVM, see
http://pyquil.readthedocs.io/en/latest/noise_models.html#support-for-noisy-gates-on-the-rigetti-qvm.
"""
            )

        self.noise_model = noise_model

        validate_noise_probabilities(gate_noise)
        validate_noise_probabilities(measurement_noise)
        self.gate_noise = gate_noise
        self.measurement_noise = measurement_noise

        if random_seed is None:
            self.random_seed = None
        elif isinstance(random_seed, int) and random_seed >= 0:
            self.random_seed = random_seed
        else:
            raise TypeError("random_seed should be None or a non-negative int")

        client_configuration = client_configuration or QCSClientConfiguration.load()
        self._qvm_client = QVMClient(client_configuration=client_configuration, request_timeout=timeout)
        self.connect()

    def connect(self) -> None:
        try:
            version = self.get_version_info()
            check_qvm_version(version)
        except ConnectionError:
            raise QVMNotRunning(f"No QVM server running at {self._qvm_client.base_url}")

    def execute(self, executable: QuantumExecutable) -> QVMExecuteResponse:
        """
        Synchronously execute the input program to completion.
        """
        executable = executable.copy()

        if not isinstance(executable, Program):
            raise TypeError(f"`QVM#executable` argument must be a `Program`; got {type(executable)}")

        result_memory = {}

        for region in executable.declarations.keys():
            result_memory[region] = np.ndarray((executable.num_shots, 0), dtype=np.int64)

        trials = executable.num_shots
        classical_addresses = get_classical_addresses_from_program(executable)

        if self.noise_model is not None:
            executable = apply_noise_model(executable, self.noise_model)

        executable._set_parameter_values_at_runtime()

        request = qvm_run_request(
            executable,
            classical_addresses,
            trials,
            self.measurement_noise,
            self.gate_noise,
            self.random_seed,
        )
        response = self._qvm_client.run_program(request)
        ram = {key: np.array(val) for key, val in response.results.items()}
        result_memory.update(ram)

        return QVMExecuteResponse(executable=executable, memory=result_memory)

    def get_result(self, execute_response: QVMExecuteResponse) -> QAMExecutionResult:
        """
        Return the results of execution on the QVM.

        Because QVM execution is synchronous, this is a no-op which returns its input.
        """
        return QAMExecutionResult(executable=execute_response.executable, readout_data=execute_response.memory)

    def get_version_info(self) -> str:
        """
        Return version information for the QVM.

        :return: String with version information
        """
        return self._qvm_client.get_version()


def validate_noise_probabilities(noise_parameter: Optional[Tuple[float, float, float]]) -> None:
    """
    Is noise_parameter a valid specification of noise probabilities for depolarizing noise?

    :param tuple noise_parameter: Tuple of 3 noise parameter values to be validated.
    """
    if not noise_parameter:
        return
    if not isinstance(noise_parameter, tuple):
        raise TypeError("noise_parameter must be a tuple")
    if any([not isinstance(value, float) for value in noise_parameter]):
        raise TypeError("noise_parameter values should all be floats")
    if len(noise_parameter) != 3:
        raise ValueError("noise_parameter tuple must be of length 3")
    if sum(noise_parameter) > 1 or sum(noise_parameter) < 0:
        raise ValueError("sum of entries in noise_parameter must be between 0 and 1 (inclusive)")
    if any([value < 0 for value in noise_parameter]):
        raise ValueError("noise_parameter values should all be non-negative")


def validate_qubit_list(qubit_list: Sequence[int]) -> Sequence[int]:
    """
    Check the validity of qubits for the payload.

    :param qubit_list: List of qubits to be validated.
    """
    if not isinstance(qubit_list, Sequence):
        raise TypeError("'qubit_list' must be of type 'Sequence'")
    if any(not isinstance(i, int) or i < 0 for i in qubit_list):
        raise TypeError("'qubit_list' must contain positive integer values")
    return qubit_list


def prepare_register_list(
    register_dict: Mapping[str, Union[bool, Sequence[int]]]
) -> Dict[str, Union[bool, Sequence[int]]]:
    """
    Canonicalize classical addresses for the payload and ready MemoryReference instances
    for serialization.

    This function will cast keys that are iterables of int-likes to a list of Python
    ints. This is to support specifying the register offsets as ``range()`` or numpy
    arrays. This mutates ``register_dict``.

    :param register_dict: The classical memory to retrieve. Specified as a dictionary:
        the keys are the names of memory regions, and the values are either (1) a list of
        integers for reading out specific entries in that memory region, or (2) True, for
        reading out the entire memory region.
    """
    if not isinstance(register_dict, dict):
        raise TypeError("register_dict must be a dict but got " + repr(register_dict))

    for k, v in register_dict.items():
        if isinstance(v, bool):
            assert v  # If boolean v must be True
            continue

        indices = [int(x) for x in v]  # support ranges, numpy, ...

        if not all(x >= 0 for x in indices):
            raise TypeError("Negative indices into classical arrays are not allowed.")
        register_dict[k] = indices

    return register_dict


def qvm_run_request(
    quil_program: Program,
    classical_addresses: Mapping[str, Union[bool, Sequence[int]]],
    trials: int,
    measurement_noise: Optional[Tuple[float, float, float]],
    gate_noise: Optional[Tuple[float, float, float]],
    random_seed: Optional[int],
) -> RunProgramRequest:
    if not quil_program:
        raise ValueError(
            "You have attempted to run an empty program."
            " Please provide gates or measure instructions to your program."
        )
    if not isinstance(quil_program, Program):
        raise TypeError("quil_program must be a Quil program object")
    classical_addresses = prepare_register_list(classical_addresses)
    if not isinstance(trials, int):
        raise TypeError("trials must be an integer")

    return RunProgramRequest(
        program=quil_program.out(calibrations=False),
        addresses=classical_addresses,  # type: ignore
        trials=trials,
        measurement_noise=measurement_noise,
        gate_noise=gate_noise,
        seed=random_seed,
    )



import logging
from math import pi
from numbers import Complex
from typing import Callable, Generator, List, Mapping, Tuple, Optional, cast

import numpy as np

from pyquil.api import QuantumComputer
from pyquil.quil import Program
from pyquil.quilatom import QubitDesignator

# import the full public API of the pyquil experiment module
from pyquil.experiment._group import (
    _max_weight_state,
    _max_weight_operator,
    construct_tpb_graph,
    group_settings as group_experiments,
    group_settings_clique_removal as group_experiments_clique_removal,
    group_settings_greedy as group_experiments_greedy,
)
from pyquil.experiment._main import (
    Experiment,
    OperatorEncoder,
)
from pyquil.experiment._result import ExperimentResult, ratio_variance
from pyquil.experiment._setting import (
    _OneQState,
    ExperimentSetting,
    SIC0,
    SIC1,
    SIC2,
    SIC3,
    TensorProductState,
    minusX,
    minusY,
    minusZ,
    plusX,
    plusY,
    plusZ,
    zeros_state,
)
from pyquil.experiment._symmetrization import SymmetrizationLevel
from pyquil.gates import RESET, RX, RY, RZ, X
from pyquil.paulis import is_identity
from pyquil.quil import Program
from pyquil.quilatom import QubitDesignator

log = logging.getLogger(__name__)


def _one_q_sic_prep(index: int, qubit: QubitDesignator) -> Program:
    """Prepare the index-th SIC basis state."""
    if index == 0:
        return Program()

    theta = 2 * np.arccos(1 / np.sqrt(3))
    zx_plane_rotation = Program([RX(-pi / 2, qubit), RZ(theta - pi, qubit), RX(-pi / 2, qubit)])

    if index == 1:
        return zx_plane_rotation

    elif index == 2:
        return zx_plane_rotation + RZ(-2 * pi / 3, qubit)

    elif index == 3:
        return zx_plane_rotation + RZ(2 * pi / 3, qubit)

    raise ValueError(f"Bad SIC index: {index}")


def _one_q_pauli_prep(label: str, index: int, qubit: QubitDesignator) -> Program:
    """Prepare the index-th eigenstate of the pauli operator given by label."""
    if index not in [0, 1]:
        raise ValueError(f"Bad Pauli index: {index}")

    if label == "X":
        if index == 0:
            return Program(RY(pi / 2, qubit))
        else:
            return Program(RY(-pi / 2, qubit))

    elif label == "Y":
        if index == 0:
            return Program(RX(-pi / 2, qubit))
        else:
            return Program(RX(pi / 2, qubit))

    elif label == "Z":
        if index == 0:
            return Program()
        else:
            return Program(RX(pi, qubit))

    raise ValueError(f"Bad Pauli label: {label}")


def _one_q_state_prep(oneq_state: _OneQState) -> Program:
    """Prepare a one qubit state.

    Either SIC[0-3], X[0-1], Y[0-1], or Z[0-1].
    """
    label = oneq_state.label
    if label == "SIC":
        return _one_q_sic_prep(oneq_state.index, oneq_state.qubit)
    elif label in ["X", "Y", "Z"]:
        return _one_q_pauli_prep(label, oneq_state.index, oneq_state.qubit)
    else:
        raise ValueError(f"Bad state label: {label}")


def _local_pauli_eig_meas(op: str, idx: QubitDesignator) -> Program:
    """
    Generate gate sequence to measure in the eigenbasis of a Pauli operator, assuming
    we are only able to measure in the Z eigenbasis. (Note: The unitary operations of this
    Program are essentially the Hermitian conjugates of those in :py:func:`_one_q_pauli_prep`)

    """
    if op == "X":
        return Program(RY(-pi / 2, idx))
    elif op == "Y":
        return Program(RX(pi / 2, idx))
    elif op == "Z":
        return Program()
    raise ValueError(f"Unknown operation {op}")


def _generate_experiment_programs(
    tomo_experiment: Experiment, active_reset: bool = False
) -> Tuple[List[Program], List[List[int]]]:
    """
    Generate the programs necessary to estimate the observables in a TomographyExperiment.
    Grouping of settings to be run in parallel, e.g. by a call to group_experiments, should be
    done before this function is called.

    .. CAUTION::
        One must be careful with compilation of the output programs before the appropriate MEASURE
        instructions are added, because compilation may re-index the qubits so that
        the output list of `measure_qubits` no longer accurately indexes the qubits that
        should be measured.

    :param tomo_experiment: a single TomographyExperiment to be translated to a series of programs
        that, when run serially, can be used to estimate each of its observables.
    :param active_reset: whether or not to begin the program by actively resetting. If true,
        execution of each of the returned programs in a loop on the QPU will generally be faster.
    :return: a list of programs along with a corresponding list of the groups of qubits that are
        measured by that program. The returned programs may be run on a qc after measurement
        instructions are added for the corresponding group of qubits in meas_qubits, or by a call
        to `qc.run_symmetrized_readout` -- see :func:`raw_estimate_observables` for possible usage.
    """
    # Outer loop over a collection of grouped settings for which we can simultaneously estimate.
    programs = []
    meas_qubits = []
    for settings in tomo_experiment:

        # Prepare a state according to the amalgam of all setting.in_state
        total_prog = Program()
        if active_reset:
            total_prog += RESET()
        max_weight_in_state = _max_weight_state(setting.in_state for setting in settings)
        if max_weight_in_state is None:
            raise ValueError(
                "Input states are not compatible. Re-group the experiment settings "
                "so that groups of parallel settings have compatible input states."
            )
        for oneq_state in max_weight_in_state.states:
            total_prog += _one_q_state_prep(oneq_state)

        # Add in the program
        total_prog += tomo_experiment.program

        # Prepare for measurement state according to setting.out_operator
        max_weight_out_op = _max_weight_operator(setting.out_operator for setting in settings)
        if max_weight_out_op is None:
            raise ValueError(
                "Observables not compatible. Re-group the experiment settings "
                "so that groups of parallel settings have compatible observables."
            )
        for qubit, op_str in max_weight_out_op:
            assert isinstance(qubit, int)
            total_prog += _local_pauli_eig_meas(op_str, qubit)

        programs.append(total_prog)

        meas_qubits.append(cast(List[int], max_weight_out_op.get_qubits()))
    return programs, meas_qubits


def measure_observables(
    qc: QuantumComputer,
    tomo_experiment: Experiment,
    progress_callback: Optional[Callable[[int, int], None]] = None,
    calibrate_readout: Optional[str] = "plus-eig",
) -> Generator[ExperimentResult, None, None]:
    """
    Measure all the observables in a TomographyExperiment.

    :param qc: A QuantumComputer which can run quantum programs
    :param tomo_experiment: A suite of tomographic observables to measure
    :param progress_callback: If not None, this function is called each time a group of
        settings is run with arguments ``f(i, len(tomo_experiment)`` such that the progress
        is ``i / len(tomo_experiment)``.
    :param calibrate_readout: Method used to calibrate the readout results. Currently, the only
        method supported is normalizing against the operator's expectation value in its +1
        eigenstate, which can be specified by setting this variable to 'plus-eig' (default value).
        The preceding symmetrization and this step together yield a more accurate estimation of
        the observable. Set to `None` if no calibration is desired.
    """
    shots = tomo_experiment.shots
    symmetrization = tomo_experiment.symmetrization
    reset = tomo_experiment.reset

    # calibration readout only works with symmetrization turned on
    if calibrate_readout is not None and symmetrization != SymmetrizationLevel.EXHAUSTIVE:
        raise ValueError(
            "Readout calibration only currently works with exhaustive readout " "symmetrization turned on."
        )

    # generate programs for each group of simultaneous settings.
    programs, meas_qubits = _generate_experiment_programs(tomo_experiment, reset)

    for i, (prog, qubits, settings) in enumerate(zip(programs, meas_qubits, tomo_experiment)):
        log.info(f"Collecting bitstrings for the {len(settings)} settings: {settings}")

        # we don't need to do any actual measurement if the combined operator is simply the
        # identity, i.e. weight=0. We handle this specially below.
        if len(qubits) > 0:
            # obtain (optionally symmetrized) bitstring results for all of the qubits
            bitstrings = qc.run_symmetrized_readout(prog, shots, symmetrization, qubits)

        if progress_callback is not None:
            progress_callback(i, len(tomo_experiment))

        # Post-process
        # Inner loop over the grouped settings. They only differ in which qubits' measurements
        # we include in the post-processing. For example, if `settings` is Z1, Z2, Z1Z2 and we
        # measure (shots, n_qubits=2) obs_strings then the full operator value involves selecting
        # either the first column, second column, or both and multiplying along the row.
        for setting in settings:
            # Get the term's coefficient so we can multiply it in later.
            coeff = setting.out_operator.coefficient
            assert isinstance(coeff, Complex)
            if not np.isclose(coeff.imag, 0):
                raise ValueError(f"{setting}'s out_operator has a complex coefficient.")
            coeff = coeff.real

            # Special case for measuring the "identity" operator, which doesn't make much
            #     sense but should happen perfectly.
            if is_identity(setting.out_operator):
                yield ExperimentResult(setting=setting, expectation=coeff, std_err=0.0, total_counts=shots)
                continue

            # Obtain statistics from result of experiment
            obs_mean, obs_var = _stats_from_measurements(
                bitstrings, {q: idx for idx, q in enumerate(qubits)}, setting, shots, coeff
            )

            if calibrate_readout == "plus-eig":
                # Readout calibration
                # Obtain calibration program
                calibr_prog = _calibration_program(qc, tomo_experiment, setting)
                calibr_qubs = setting.out_operator.get_qubits()
                calibr_qub_dict = {cast(int, q): idx for idx, q in enumerate(calibr_qubs)}

                # Perform symmetrization on the calibration program
                calibr_results = qc.run_symmetrized_readout(
                    calibr_prog, shots, SymmetrizationLevel.EXHAUSTIVE, calibr_qubs
                )

                # Obtain statistics from the measurement process
                obs_calibr_mean, obs_calibr_var = _stats_from_measurements(
                    calibr_results, calibr_qub_dict, setting, shots
                )
                # Calibrate the readout results
                corrected_mean = obs_mean / obs_calibr_mean
                corrected_var = ratio_variance(obs_mean, obs_var, obs_calibr_mean, obs_calibr_var)

                yield ExperimentResult(
                    setting=setting,
                    expectation=corrected_mean.item(),
                    std_err=np.sqrt(corrected_var).item(),
                    total_counts=len(bitstrings),
                    raw_expectation=obs_mean.item(),
                    raw_std_err=np.sqrt(obs_var).item(),
                    calibration_expectation=obs_calibr_mean.item(),
                    calibration_std_err=np.sqrt(obs_calibr_var).item(),
                    calibration_counts=len(calibr_results),
                )

            elif calibrate_readout is None:
                # No calibration
                yield ExperimentResult(
                    setting=setting,
                    expectation=obs_mean.item(),
                    std_err=np.sqrt(obs_var).item(),
                    total_counts=len(bitstrings),
                )

            else:
                raise ValueError("Calibration readout method must be either 'plus-eig' or None")


def _ops_bool_to_prog(ops_bool: Tuple[bool], qubits: List[int]) -> Program:
    """
    :param ops_bool: tuple of booleans specifying the operation to be carried out on `qubits`
    :param qubits: list specifying the qubits to be carried operations on
    :return: Program with the operations specified in `ops_bool` on the qubits specified in
        `qubits`
    """
    assert len(ops_bool) == len(qubits), "Mismatch of qubits and operations"
    prog = Program()
    for i, op_bool in enumerate(ops_bool):
        if op_bool == 0:
            continue
        elif op_bool == 1:
            prog += Program(X(qubits[i]))
        else:
            raise ValueError("ops_bool should only consist of 0s and/or 1s")
    return prog


def _stats_from_measurements(
    bs_results: np.ndarray,
    qubit_index_map: Mapping[int, int],
    setting: ExperimentSetting,
    n_shots: int,
    coeff: float = 1.0,
) -> Tuple[np.number, np.number]:  # type: ignore
    """
    :param bs_results: results from running `qc.run`
    :param qubit_index_map: dict mapping qubit to classical register index
    :param setting: ExperimentSetting
    :param n_shots: number of shots in the measurement process
    :param coeff: coefficient of the operator being estimated
    :return: tuple specifying (mean, variance)
    """
    # Identify classical register indices to select
    idxs = [qubit_index_map[cast(int, q)] for q, _ in setting.out_operator]
    # Pick columns corresponding to qubits with a non-identity out_operation
    obs_strings = bs_results[:, idxs]
    # Transform bits to eigenvalues; ie (+1, -1)
    my_obs_strings = 1 - 2 * obs_strings
    # Multiply row-wise to get operator values. Do statistics. Return result.
    obs_vals = coeff * np.prod(my_obs_strings, axis=1)
    obs_mean = np.mean(obs_vals)
    obs_var = np.var(obs_vals) / n_shots

    return obs_mean, obs_var


def _calibration_program(qc: QuantumComputer, tomo_experiment: Experiment, setting: ExperimentSetting) -> Program:
    """
    Program required for calibration in a tomography-like experiment.

    :param tomo_experiment: A suite of tomographic observables
    :param ExperimentSetting: The particular tomographic observable to measure
    :param symmetrize_readout: Method used to symmetrize the readout errors (see docstring for
        `measure_observables` for more details)
    :param cablir_shots: number of shots to take in the measurement process
    :return: Program performing the calibration
    """
    # Inherit any noisy attributes from main Program, including gate definitions
    # and applications which can be handy in creating simulating noisy channels
    calibr_prog = Program()
    # Inherit readout errro instructions from main Program
    readout_povm_instruction = [i for i in tomo_experiment.program.out().split("\n") if "PRAGMA READOUT-POVM" in i]
    calibr_prog += readout_povm_instruction
    # Inherit any definitions of noisy gates from main Program
    kraus_instructions = [i for i in tomo_experiment.program.out().split("\n") if "PRAGMA ADD-KRAUS" in i]
    calibr_prog += kraus_instructions
    # Prepare the +1 eigenstate for the out operator
    for q, op in setting.out_operator.operations_as_set():
        assert isinstance(q, int)
        calibr_prog += _one_q_pauli_prep(label=op, index=0, qubit=q)
    # Measure the out operator in this state
    for q, op in setting.out_operator.operations_as_set():
        assert isinstance(q, int)
        calibr_prog += _local_pauli_eig_meas(op, q)

    return calibr_prog



from typing import Any
from pyquil.api._qvm import QVM as QVMV3
from ._qam import StatefulQAM


class QVM(QVMV3):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        StatefulQAM.wrap(self)



##############################################################################
# Copyright 2019 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

from collections import defaultdict
from dataclasses import dataclass, replace
from typing import Iterable, List, Sequence, Mapping, Optional, Set, Tuple, cast
from warnings import warn

from pyquil.quil import Program
from pyquil.quilatom import ParameterDesignator, QubitDesignator, format_parameter
from pyquil.quilbase import (
    AbstractInstruction,
    Wait,
    ResetQubit,
    JumpConditional,
    JumpWhen,
    JumpUnless,
    Jump,
    UnaryClassicalInstruction,
    LogicalBinaryOp,
    ArithmeticBinaryOp,
    ClassicalMove,
    ClassicalExchange,
    ClassicalConvert,
    ClassicalLoad,
    ClassicalStore,
    ClassicalComparison,
    RawInstr,
    Measurement,
    Gate,
    Pragma,
)


@dataclass
class DiagramSettings:
    """
    Settings to control the layout and rendering of circuits.
    """

    texify_numerical_constants: bool = True
    """
    Convert numerical constants, such as pi, to LaTeX form.
    """

    impute_missing_qubits: bool = False
    """
    Include qubits with indices between those explicitly referenced in the Quil program.

    For example, if true, the diagram for `CNOT 0 2` would have three qubit lines: 0, 1, 2.
    """

    label_qubit_lines: bool = True
    """
    Label qubit lines.
    """

    abbreviate_controlled_rotations: bool = False
    """
    Write controlled rotations in a compact form.

    For example,  `RX(pi)` as `X_{\\pi}`, instead of the longer `R_X(\\pi)`
    """

    qubit_line_open_wire_length: int = 1
    """
    The length by which qubit lines should be extended with open wires at the right of the diagram.

    The default of 1 is the natural choice. The main reason for including this option
    is that it may be appropriate for this to be 0 in subdiagrams.
    """

    right_align_terminal_measurements: bool = True
    """
    Align measurement operations which appear at the end of the program.
    """


# Constants


PRAGMA_BEGIN_GROUP = "LATEX_GATE_GROUP"
PRAGMA_END_GROUP = "END_LATEX_GATE_GROUP"

UNSUPPORTED_INSTRUCTION_CLASSES = (
    Wait,
    JumpConditional,
    JumpWhen,
    JumpUnless,
    Jump,
    UnaryClassicalInstruction,
    LogicalBinaryOp,
    ArithmeticBinaryOp,
    ClassicalMove,
    ClassicalExchange,
    ClassicalConvert,
    ClassicalLoad,
    ClassicalStore,
    ClassicalComparison,
    RawInstr,
)


# TikZ operators


def TIKZ_LEFT_KET(qubit: int) -> str:
    return r"\lstick{{\ket{{q_{{{qubit}}}}}}}".format(qubit=qubit)


def TIKZ_CONTROL(control: int, offset: int) -> str:
    return r"\ctrl{{{offset}}}".format(offset=offset)


def TIKZ_CNOT_TARGET() -> str:
    return r"\targ{}"


def TIKZ_CPHASE_TARGET() -> str:
    return r"\control{}"


def TIKZ_SWAP(source: int, offset: int) -> str:
    return r"\swap{{{offset}}}".format(offset=offset)


def TIKZ_SWAP_TARGET() -> str:
    return r"\targX{}"


def TIKZ_NOP() -> str:
    return r"\qw"


def TIKZ_MEASURE() -> str:
    return r"\meter{}"


def _format_parameter(param: ParameterDesignator, settings: Optional[DiagramSettings] = None) -> str:
    formatted = format_parameter(param)
    if settings and settings.texify_numerical_constants:
        formatted = formatted.replace("pi", r"\pi")
    return formatted


def _format_parameters(params: Iterable[ParameterDesignator], settings: Optional[DiagramSettings] = None) -> str:
    return "(" + ",".join(_format_parameter(param, settings) for param in params) + ")"


def TIKZ_GATE(
    name: str,
    size: int = 1,
    params: Optional[Sequence[ParameterDesignator]] = None,
    dagger: bool = False,
    settings: Optional[DiagramSettings] = None,
) -> str:
    cmd = r"\gate"
    rotations = ["RX", "RY", "RZ"]
    if settings and settings.abbreviate_controlled_rotations and name in rotations and params:
        name = name[1] + "_{{{param}}}".format(param=_format_parameter(params[0], settings))
        return cmd + "{{{name}}}".format(name=name)
    # now, handle the general case
    if size > 1:
        cmd += "[wires={size}]".format(size=size)
    # TeXify names
    if name in ["RX", "RY", "RZ"]:
        name = name[0] + "_" + name[1].lower()
    if dagger:
        name += r"^{\dagger}"
    if params:
        name += _format_parameters(params, settings)
    return cmd + "{{{name}}}".format(name=name)


def TIKZ_GATE_GROUP(qubits: Sequence[int], width: int, label: str) -> str:
    num_qubits = max(qubits) - min(qubits) + 1
    return (
        "\\gategroup[{qubits},steps={width},style={{dashed, rounded corners,"
        "fill=blue!20, inner xsep=2pt}}, background]{{{label}}}".format(qubits=num_qubits, width=width, label=label)
    )


SOURCE_TARGET_OP = {
    "CNOT": (TIKZ_CONTROL, TIKZ_CNOT_TARGET),
    "SWAP": (TIKZ_SWAP, TIKZ_SWAP_TARGET),
    "CZ": (TIKZ_CONTROL, lambda: TIKZ_GATE("Z")),
    "CPHASE": (TIKZ_CONTROL, TIKZ_CPHASE_TARGET),
}


class DiagramState:
    """
    A representation of a circuit diagram.

    This maintains an ordered list of qubits, and for each qubit a 'line': that is, a list of
    TikZ operators.
    """

    def __init__(self, qubits: Sequence[int]):
        self.qubits = qubits
        self.lines: Mapping[int, List[str]] = defaultdict(list)

    def extend_lines_to_common_edge(self, qubits: Iterable[int], offset: int = 0) -> None:
        """
        Add NOP operations on the lines associated with the given qubits, until
        all lines are of the same width.
        """
        max_width = max(self.width(q) for q in qubits) + offset
        for q in qubits:
            while self.width(q) < max_width:
                self.append(q, TIKZ_NOP())

    def width(self, qubit: int) -> int:
        """
        The width of the diagram, in terms of the number of operations, on the
        specified qubit line.
        """
        return len(self.lines[qubit])

    def append(self, qubit: int, op: str) -> None:
        """
        Add an operation to the rightmost edge of the specified qubit line.
        """
        self.lines[qubit].append(op)

    def append_diagram(self, diagram: "DiagramState", group: Optional[str] = None) -> "DiagramState":
        """
        Add all operations represented by the given diagram to their
        corresponding qubit lines in this diagram.

        If group is not None, then a TIKZ_GATE_GROUP is created with the label indicated by group.
        """
        grouped_qubits = diagram.qubits
        diagram.extend_lines_to_common_edge(grouped_qubits)
        # NOTE: this may create new lines, no big deal
        self.extend_lines_to_common_edge(grouped_qubits)

        # record info for later (optional) group placement
        # the group is marked with a rectangle. we compute the upper-left corner and the width of
        # the rectangle
        corner_row = grouped_qubits[0]
        corner_col = len(self.lines[corner_row]) + 1
        group_width = diagram.width(corner_row) - 1

        # append ops to this diagram
        for q in diagram.qubits:
            for op in diagram.lines[q]:
                self.append(q, op)
        # add tikz grouping command
        if group is not None:
            self.lines[corner_row][corner_col] += " " + TIKZ_GATE_GROUP(grouped_qubits, group_width, group)
        return self

    def interval(self, low: int, high: int) -> List[int]:
        """
        All qubits in the diagram, from low to high, inclusive.
        """
        full_interval = range(low, high + 1)
        qubits = list(set(full_interval) & set(self.qubits))
        return sorted(qubits)

    def is_interval(self, qubits: Sequence[int]) -> bool:
        """
        Do the specified qubits correspond to an interval in this diagram?
        """
        return qubits == self.interval(min(qubits), max(qubits))


def split_on_terminal_measures(
    program: Program,
) -> Tuple[List[AbstractInstruction], List[AbstractInstruction]]:
    """
    Split a program into two lists of instructions:

    1. A set of measurement instructions occuring as the final operation on their qubit.
    2. The rest.
    """
    # handle the easy case explicitly (mainly to avoid warning when we can avoid it)
    if not any(isinstance(instr, Measurement) for instr in program.instructions):
        return [], program.instructions

    seen_qubits: Set[QubitDesignator] = set()

    measures: List[AbstractInstruction] = []
    remaining: List[AbstractInstruction] = []
    in_group = False
    for instr in reversed(program.instructions):
        if not in_group and isinstance(instr, Measurement) and instr.qubit not in seen_qubits:
            measures.insert(0, instr)
            seen_qubits.add(instr.qubit)
        else:
            remaining.insert(0, instr)
            if isinstance(instr, (Gate, ResetQubit)):
                seen_qubits |= instr.get_qubits()
            elif isinstance(instr, Pragma):
                if instr.command == PRAGMA_END_GROUP:
                    warn("Alignment of terminal MEASURE operations may" "conflict with gate group declaration.")
                    in_group = True
                elif instr.command == PRAGMA_BEGIN_GROUP:
                    in_group = False
    return measures, remaining


class DiagramBuilder:
    """
    Constructs DiagramStates from a given circuit and settings.

    This is essentially a state machine, represented by a few instance variables and some mutually
    recursive methods.
    """

    def __init__(self, circuit: Program, settings: DiagramSettings):
        self.circuit = circuit
        self.settings = settings
        # instructions currently being processed
        self.working_instructions: Optional[List[AbstractInstruction]] = None
        # index into working instructions. we maintain the invariant that
        # working_instructions[0:index] has been processed, with the diagram
        # updated accordingly
        self.index = 0
        # partially constructed diagram
        self.diagram: Optional[DiagramState] = None

    def build(self) -> DiagramState:
        """
        Actually build the diagram.
        """
        qubits = cast(Set[int], self.circuit.get_qubits(indices=True))
        all_qubits = range(min(qubits), max(qubits) + 1) if self.settings.impute_missing_qubits else sorted(qubits)
        self.diagram = DiagramState(all_qubits)

        if self.settings.right_align_terminal_measurements:
            measures, instructions = split_on_terminal_measures(self.circuit)
        else:
            measures, instructions = [], self.circuit.instructions

        # setup the left fringe
        if self.settings.label_qubit_lines:
            for qubit in self.diagram.qubits:
                self.diagram.append(qubit, TIKZ_LEFT_KET(qubit))
        else:  # initial exposed wires
            self.diagram.extend_lines_to_common_edge(self.diagram.qubits, offset=1)

        # setup working state
        self.working_instructions = instructions
        self.index = 0

        # main loop
        while self.index < len(self.working_instructions):
            instr = self.working_instructions[self.index]
            if isinstance(instr, Pragma) and instr.command == PRAGMA_BEGIN_GROUP:
                self._build_group()
            elif isinstance(instr, Pragma) and instr.command == PRAGMA_END_GROUP:
                raise ValueError("PRAGMA {} found without matching {}.".format(PRAGMA_END_GROUP, PRAGMA_BEGIN_GROUP))
            elif isinstance(instr, Measurement):
                self._build_measure()
            elif isinstance(instr, Gate):
                if "FORKED" in instr.modifiers:
                    raise ValueError("LaTeX output does not currently support" "FORKED modifiers: {}.".format(instr))
                # the easy case is 1q operations
                if len(instr.qubits) == 1:
                    self._build_1q_unitary()
                else:
                    if instr.name in SOURCE_TARGET_OP and not instr.modifiers:
                        self._build_custom_source_target_op()
                    else:
                        self._build_generic_unitary()
            elif isinstance(instr, UNSUPPORTED_INSTRUCTION_CLASSES):
                raise ValueError(
                    "LaTeX output does not currently support" "the following instruction: {}".format(instr.out())
                )
            else:
                self.index += 1

        self.diagram.extend_lines_to_common_edge(self.diagram.qubits)

        # handle terminal measurements
        self.index = 0
        self.working_instructions = measures

        for _ in self.working_instructions:
            self._build_measure()

        offset = max(self.settings.qubit_line_open_wire_length, 0)
        self.diagram.extend_lines_to_common_edge(self.diagram.qubits, offset=offset)
        return self.diagram

    def _build_group(self) -> None:
        """
        Update the partial diagram with the subcircuit delimited by the grouping PRAGMA.

        Advances the index beyond the ending pragma.
        """
        assert self.working_instructions is not None
        instr = self.working_instructions[self.index]
        assert isinstance(instr, Pragma)
        if len(instr.args) != 0:
            raise ValueError(f"PRAGMA {PRAGMA_BEGIN_GROUP} expected a freeform string, or nothing at all.")
        start = self.index + 1
        # walk instructions until the group end
        for j in range(start, len(self.working_instructions)):
            instruction_j = self.working_instructions[j]
            if isinstance(instruction_j, Pragma) and instruction_j.command == PRAGMA_END_GROUP:
                # recursively build the diagram for this block
                # we do not want labels here!
                block_settings = replace(self.settings, label_qubit_lines=False, qubit_line_open_wire_length=0)
                subcircuit = Program(*self.working_instructions[start:j])
                block = DiagramBuilder(subcircuit, block_settings).build()
                block_name = instr.freeform_string if instr.freeform_string else ""
                assert self.diagram is not None
                self.diagram.append_diagram(block, group=block_name)
                # advance to the instruction following this one
                self.index = j + 1
                return

        raise ValueError("Unable to find PRAGMA {} matching {}.".format(PRAGMA_END_GROUP, instr))

    def _build_measure(self) -> None:
        """
        Update the partial diagram with a measurement operation.

        Advances the index by one.
        """
        assert self.working_instructions is not None
        instr = self.working_instructions[self.index]
        assert isinstance(instr, Measurement)
        assert self.diagram is not None
        self.diagram.append(instr.qubit.index, TIKZ_MEASURE())
        self.index += 1

    def _build_custom_source_target_op(self) -> None:
        """
        Update the partial diagram with a single operation involving a source and a target
        (e.g. a controlled gate, a swap).

        Advances the index by one.
        """
        assert self.working_instructions is not None
        instr = self.working_instructions[self.index]
        assert isinstance(instr, Gate)
        source, target = qubit_indices(instr)
        assert self.diagram is not None
        displaced = self.diagram.interval(min(source, target), max(source, target))
        self.diagram.extend_lines_to_common_edge(displaced)
        source_op, target_op = SOURCE_TARGET_OP[instr.name]
        offset = (-1 if source > target else 1) * (len(displaced) - 1)  # a directed quantity
        self.diagram.append(source, source_op(source, offset))
        self.diagram.append(target, target_op())
        self.diagram.extend_lines_to_common_edge(displaced)
        self.index += 1

    def _build_1q_unitary(self) -> None:
        """
        Update the partial diagram with a 1Q gate.

        Advances the index by one.
        """
        assert self.working_instructions is not None
        instr = self.working_instructions[self.index]
        assert isinstance(instr, Gate)
        qubits = qubit_indices(instr)
        dagger = sum(m == "DAGGER" for m in instr.modifiers) % 2 == 1

        assert self.diagram is not None
        self.diagram.append(
            qubits[0],
            TIKZ_GATE(instr.name, params=instr.params, dagger=dagger, settings=self.settings),
        )
        self.index += 1

    def _build_generic_unitary(self) -> None:
        """
        Update the partial diagram with a unitary operation.

        Advances the index by one.
        """
        assert self.working_instructions is not None
        instr = self.working_instructions[self.index]
        assert isinstance(instr, Gate)
        qubits = qubit_indices(instr)
        dagger = sum(m == "DAGGER" for m in instr.modifiers) % 2 == 1
        controls = sum(m == "CONTROLLED" for m in instr.modifiers)

        assert self.diagram is not None
        self.diagram.extend_lines_to_common_edge(qubits)

        control_qubits = qubits[:controls]
        target_qubits = qubits[controls:]
        if not self.diagram.is_interval(sorted(target_qubits)):
            raise ValueError(f"Unable to render instruction {instr} which targets non-adjacent qubits.")

        for q in control_qubits:
            offset = target_qubits[0] - q
            self.diagram.append(q, TIKZ_CONTROL(q, offset))

        # we put the gate on the first target line, and nop on the others
        self.diagram.append(
            target_qubits[0],
            TIKZ_GATE(instr.name, size=len(target_qubits), params=instr.params, dagger=dagger),
        )
        for q in target_qubits[1:]:
            self.diagram.append(q, TIKZ_NOP())

        self.index += 1


def qubit_indices(instr: AbstractInstruction) -> List[int]:
    """
    Get a list of indices associated with the given instruction.
    """
    if isinstance(instr, Measurement):
        return [instr.qubit.index]
    elif isinstance(instr, Gate):
        return [qubit.index for qubit in instr.qubits]
    else:
        return []



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyquil',
 'pyquil._parser',
 'pyquil.api',
 'pyquil.compatibility',
 'pyquil.compatibility.v2',
 'pyquil.compatibility.v2.api',
 'pyquil.experiment',
 'pyquil.external',
 'pyquil.latex',
 'pyquil.quantum_processor',
 'pyquil.quantum_processor.transformers',
 'pyquil.simulation']

package_data = \
{'': ['*']}

install_requires = \
['lark>=0.11.1,<0.12.0',
 'networkx>=2.5,<3.0',
 'numpy>=1.20,<2.0',
 'qcs-api-client>=0.20.13,<0.22.0',
 'retry>=0.9.2,<0.10.0',
 'rpcq>=3.10.0,<4.0.0',
 'scipy>=1.6.1,<2.0.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=3.7.3,<4.0.0'],
 'docs': ['Sphinx>=4.0.2,<5.0.0',
          'sphinx-rtd-theme>=0.5.2,<0.6.0',
          'nbsphinx>=0.8.6,<0.9.0',
          'recommonmark>=0.7.1,<0.8.0'],
 'latex': ['ipython>=7.21.0,<8.0.0']}

setup_kwargs = {
    'name': 'pyquil',
    'version': '3.2.0rc7',
    'description': 'A Python library for creating Quantum Instruction Language (Quil) programs.',
    'long_description': 'PyQuil: Quantum programming in Python\n=====================================\n\n[![binder](https://mybinder.org/badge_logo.svg)][binder]\n[![docs][docs-badge]][docs-repo]\n[![coverage][coverage-badge]][coverage-repo]\n[![docker][docker-badge]][docker-repo]\n[![pepy][pepy-badge]][pepy-repo]\n[![pypi][pypi-badge]][pypi-repo]\n[![slack][slack-badge]][slack-invite]\n\nPyQuil is a Python library for quantum programming using [Quil](https://arxiv.org/abs/1608.03355),\nthe quantum instruction language developed at [Rigetti Computing](https://www.rigetti.com/).\nPyQuil serves three main functions:\n\n- Easily generating Quil programs from quantum gates and classical operations\n- Compiling and simulating Quil programs using the [Quil Compiler](https://github.com/rigetti/quilc)\n  (quilc) and the [Quantum Virtual Machine](https://github.com/rigetti/qvm) (QVM)\n- Executing Quil programs on real quantum processors (QPUs) using\n  [Quantum Cloud Services][qcs-paper] (QCS)\n\nPyQuil has a ton of other features, which you can learn more about in the\n[docs](http://pyquil.readthedocs.io/en/latest/). However, you can also keep reading\nbelow to get started with running your first quantum program!\n\nQuickstart with interactive tutorial notebooks\n----------------------------------------------\n\nWithout installing anything, you can quickly get started with quantum programming by exploring\nour interactive [Jupyter][jupyter] notebook tutorials and examples. To run them in a preconfigured\nexecution environment on [Binder][mybinder], click the "launch binder" badge at the top of the\nREADME or the link [here][binder]! To learn more about the tutorials and how you can add your own,\nvisit the [rigetti/forest-tutorials][forest-tutorials] repository. If you\'d rather set everything\nup locally, or are interested in contributing to pyQuil, continue onto the next section for\ninstructions on installing pyQuil and the Forest SDK.\n\nInstalling pyQuil and the Forest SDK\n------------------------------------\n\n[![pypi][pypi-badge]][pypi-repo]\n[![conda-forge][conda-forge-badge]][conda-forge-badge]\n[![conda-rigetti][conda-rigetti-badge]][conda-rigetti-repo]\n\nPyQuil can be installed using `conda`, `pip`, or from source. To install it from PyPI (via `pip`),\ndo the following:\n\n```bash\npip install pyquil\n```\n\nTo instead install pyQuil from source, do the following from within the repository after cloning it:\n\n```bash\npip install -e .\n```\n\nIf you choose to use `pip`, we highly recommend installing pyQuil within a virtual environment.\n\nPyQuil, along with quilc, the QVM, and other libraries, make up what is called the Forest\nSDK. To make full use of pyQuil, you will need to additionally have installed\n[quilc](https://github.com/quil-lang/quilc) and the [QVM](https://github.com/quil-lang/qvm).\nFor more information, check out the docs!\n\nRunning your first quantum program\n----------------------------------\n\nIn just a few lines, we can use pyQuil with the Forest SDK to simulate a Bell state!\n\n```python\nfrom pyquil import get_qc, Program\nfrom pyquil.gates import CNOT, H, MEASURE\n \nqvm = get_qc(\'2q-qvm\')\n \np = Program()\np += H(0)\np += CNOT(0, 1)\nro = p.declare(\'ro\', \'BIT\', 2)\np += MEASURE(0, ro[0])\np += MEASURE(1, ro[1])\np.wrap_in_numshots_loop(10)\n\nqvm.run(p).readout_data[\'ro\'].tolist()\n```\n\nThe output of the above program should look something like the following,\nthe statistics of which are consistent with a two-qubit entangled state.\n\n```\n[[0, 0],\n [1, 1],\n [1, 1],\n [1, 1],\n [1, 1],\n [0, 0],\n [0, 0],\n [1, 1],\n [0, 0],\n [0, 0]]\n```\n\nUsing the Forest SDK, you can simulate the operation of a real quantum processor (QPU). If you\nwould like to run on the real QPUs in our lab in Berkeley, you can sign up for an account\non [Quantum Cloud Services][qcs-request-access] (QCS)!\n\nJoining the Forest community\n----------------------------\n\nIf you\'d like to get involved with pyQuil and Forest, joining the\n[Rigetti Forest Slack Workspace][slack-invite] is a great place to start! You can do so by\nclicking the invite link in the previous sentence, or in the badge at the top of this README.\nThe Slack Workspace is a great place to ask general questions, join high-level design discussions,\nand hear about updates to pyQuil and the Forest SDK.\n\nTo go a step further and start contributing to the development of pyQuil, good first steps are\n[reporting a bug][bug], [requesting a feature][feature], or picking up one of the issues with the\n[good first issue][first] or [help wanted][help] labels. Once you find an issue to work\non, make sure to [fork this repository][fork] and then [open a pull request][pr] once your changes\nare ready. For more information on all the ways you can contribute to pyQuil (along with\nsome helpful tips for developers and maintainers) check out our\n[Contributing Guide](CONTRIBUTING.md)!\n\nTo see what people have contributed in the past, check out the [Changelog](CHANGELOG.md) for\na detailed list of all announcements, improvements, changes, and bugfixes. The\n[Releases](https://github.com/rigetti/pyquil/releases) page for pyQuil contains similar\ninformation, but with links to the pull request for each change and its corresponding author.\nThanks for contributing to pyQuil! 🙂\n\nCiting pyQuil, Forest, and Quantum Cloud Services\n-------------------------------------------------\n\n[![zenodo][zenodo-badge]][zenodo-doi]\n\nIf you use pyQuil, Grove, or other parts of the Forest SDK in your research, please cite\nthe [Quil specification][quil-paper] using the following BibTeX snippet:\n\n```bibtex\n@misc{smith2016practical,\n    title={A Practical Quantum Instruction Set Architecture},\n    author={Robert S. Smith and Michael J. Curtis and William J. Zeng},\n    year={2016},\n    eprint={1608.03355},\n    archivePrefix={arXiv},\n    primaryClass={quant-ph}\n}\n```\n\nAdditionally, if your research involves taking data on Rigetti quantum processors (QPUs) via\nthe Quantum Cloud Services (QCS) platform, please reference the [QCS paper][qcs-paper] using the\nfollowing BibTeX snippet:\n\n```bibtex\n@article{Karalekas_2020,\n    title = {A quantum-classical cloud platform optimized for variational hybrid algorithms},\n    author = {Peter J Karalekas and Nikolas A Tezak and Eric C Peterson\n              and Colm A Ryan and Marcus P da Silva and Robert S Smith},\n    year = 2020,\n    month = {apr},\n    publisher = {{IOP} Publishing},\n    journal = {Quantum Science and Technology},\n    volume = {5},\n    number = {2},\n    pages = {024003},\n    doi = {10.1088/2058-9565/ab7559},\n    url = {https://doi.org/10.1088%2F2058-9565%2Fab7559},\n}\n```\n\nThe preprint of the QCS paper is available on [arXiv][qcs-arxiv], and the supplementary\ninteractive notebooks and datasets for the paper can be found in the [rigetti/qcs-paper][qcs-repo]\nrepository.\n\nLicense\n-------\n\nPyQuil is licensed under the\n[Apache License 2.0](https://github.com/rigetti/pyQuil/blob/master/LICENSE).\n\n[binder]: https://mybinder.org/v2/gh/rigetti/forest-tutorials/master?urlpath=lab/tree/Welcome.ipynb\n[conda-forge-badge]: https://img.shields.io/conda/vn/conda-forge/pyquil.svg\n[conda-forge-repo]: https://anaconda.org/conda-forge/pyquil\n[conda-rigetti-badge]: https://img.shields.io/conda/vn/rigetti/pyquil?label=conda-rigetti\n[conda-rigetti-repo]: https://anaconda.org/rigetti/pyquil\n[coverage-badge]: https://coveralls.io/repos/github/rigetti/pyquil/badge.svg?branch=more-badges\n[coverage-repo]: https://coveralls.io/github/rigetti/pyquil?branch=more-badges\n[docker-badge]: https://img.shields.io/docker/pulls/rigetti/forest\n[docker-repo]: https://hub.docker.com/r/rigetti/forest\n[docs-badge]: https://readthedocs.org/projects/pyquil/badge/?version=latest\n[docs-repo]: http://pyquil.readthedocs.io/en/latest/?badge=latest\n[forest-tutorials]: https://github.com/rigetti/forest-tutorials\n[jupyter]: https://jupyter.org/\n[mybinder]: https://mybinder.org\n[pepy-badge]: https://pepy.tech/badge/pyquil\n[pepy-repo]: https://pepy.tech/project/pyquil\n[pypi-badge]: https://img.shields.io/pypi/v/pyquil.svg\n[pypi-repo]: https://pypi.org/project/pyquil/\n[qcs-request-access]: https://qcs.rigetti.com/request-access\n[slack-badge]: https://img.shields.io/badge/slack-rigetti--forest-812f82.svg?\n[zenodo-badge]: https://zenodo.org/badge/DOI/10.5281/zenodo.3553165.svg\n[zenodo-doi]: https://doi.org/10.5281/zenodo.3553165\n\n[qcs-arxiv]: https://arxiv.org/abs/2001.04449\n[qcs-paper]: https://dx.doi.org/10.1088/2058-9565/ab7559\n[qcs-repo]: https://github.com/rigetti/qcs-paper\n[quil-paper]: https://arxiv.org/abs/1608.03355\n\n[bug]: https://github.com/rigetti/pyquil/issues/new?assignees=&labels=bug+%3Abug%3A&template=BUG_REPORT.md&title=\n[feature]: https://github.com/rigetti/pyquil/issues/new?assignees=&labels=enhancement+%3Asparkles%3A&template=FEATURE_REQUEST.md&title=\n[first]: https://github.com/rigetti/pyquil/labels/good%20first%20issue%20%3Ababy%3A\n[help]: https://github.com/rigetti/pyquil/labels/help%20wanted%20%3Awave%3A\n[fork]: https://github.com/rigetti/pyquil/fork\n[pr]: https://github.com/rigetti/pyquil/compare\n[slack-invite]: https://join.slack.com/t/rigetti-forest/shared_invite/enQtNTUyNTE1ODg3MzE2LWQwNzBlMjZlMmNlN2M5MzQyZDlmOGViODQ5ODI0NWMwNmYzODY4YTc2ZjdjOTNmNzhiYTk2YjVhNTE2NTRkODY\n',
    'author': 'Rigetti Computing',
    'author_email': 'softapps@rigetti.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rigetti/pyquil.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



##############################################################################
# Copyright 2016-2019 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

import itertools
from typing import Dict, List, Tuple, cast

import numpy as np

from pyquil.paulis import PauliTerm


def euler_angles_RX(theta: float) -> Tuple[float, float, float]:
    """
    A tuple of angles which corresponds to a ZXZXZ-decomposed ``RX`` gate.

    :param theta: The angle parameter for the ``RX`` gate.
    :return: The corresponding Euler angles for that gate.
    """
    return (np.pi / 2, theta, -np.pi / 2)


def euler_angles_RY(theta: float) -> Tuple[float, float, float]:
    """
    A tuple of angles which corresponds to a ZXZXZ-decomposed ``RY`` gate.

    :param theta: The angle parameter for the ``RY`` gate.
    :return: The corresponding Euler angles for that gate.
    """
    return (0.0, theta, 0.0)


# euler angles for preparing the +1 eigenstate of X, Y, or Z
P_X = euler_angles_RY(np.pi / 2)
P_Y = euler_angles_RX(-np.pi / 2)
P_Z = (0.0, 0.0, 0.0)


# euler angles for measuring in the eigenbasis of X, Y, or Z
M_X = euler_angles_RY(-np.pi / 2)
M_Y = euler_angles_RX(np.pi / 2)
M_Z = (0.0, 0.0, 0.0)


def pauli_term_to_euler_memory_map(
    term: PauliTerm,
    *,
    prefix: str,
    tuple_x: Tuple[float, float, float],
    tuple_y: Tuple[float, float, float],
    tuple_z: Tuple[float, float, float],
    suffix_alpha: str = "alpha",
    suffix_beta: str = "beta",
    suffix_gamma: str = "gamma",
) -> Dict[str, List[float]]:
    """
    Given a ``PauliTerm``, create a memory map corresponding to a collection of ZXZXZ-decomposed
    single-qubit gates. The intent is that these gate are used to prepare an eigenstate of the
    ``PauliTerm`` or measure in the eigenbasis of the ``PauliTerm``, which is more clearly
    discernible from the calling functions ``pauli_term_to_preparation_memory_map`` (for state
    preparation) and ``pauli_term_to_measurement_memory_map`` (for measuring in different bases).
    This function is not really meant to be used by itself, but rather by the aforementioned
    calling functions.

    :param term: The ``PauliTerm`` in question.
    :param prefix: The prefix for the declared memory region labels. For example, if the prefix
        is "preparation" and the alpha, beta, and gamma suffixes are left as default, the labels
        would be "preparation_alpha", "preparation_beta", and "preparation_gamma".
    :param tuple_x: A tuple of Euler angles as (alpha, beta, gamma) to be used for the ``X``
        operators in the ``PauliTerm``.
    :param tuple_y: A tuple of Euler angles as (alpha, beta, gamma) to be used for the ``Y``
        operators in the ``PauliTerm``.
    :param tuple_z: A tuple of Euler angles as (alpha, beta, gamma) to be used for the ``Z``
        and ``I`` operators in the ``PauliTerm``.
    :param suffix_alpha: The suffix for the "alpha" memory region label, which corresponds to the
        first (rightmost) ``Z`` in the ZXZXZ decomposition. Defaults to "alpha".
    :param suffix_beta: The suffix for the "beta" memory region label, which corresponds to the
        second (middle) ``Z`` in the ZXZXZ decomposition. Defaults to "beta".
    :param suffix_gamma: The suffix for the "gamma" memory region label, which corresponds to the
        last (leftmost) ``Z`` in the ZXZXZ decomposition. Defaults to "gamma".
    :return: Memory map dictionary containing three entries (three labels as keys and three lists
        of angles as values).
    """
    # no need to provide a memory map when no rotations are necessary
    if ("X" not in term.pauli_string(cast(List[int], term.get_qubits()))) and (
        "Y" not in term.pauli_string(cast(List[int], term.get_qubits()))
    ):
        return {}

    alpha_label = f"{prefix}_{suffix_alpha}"
    beta_label = f"{prefix}_{suffix_beta}"
    gamma_label = f"{prefix}_{suffix_gamma}"

    # assume the pauli indices are equivalent to the memory region
    memory_size = max(cast(List[int], term.get_qubits())) + 1

    memory_map = {
        alpha_label: [0.0] * memory_size,
        beta_label: [0.0] * memory_size,
        gamma_label: [0.0] * memory_size,
    }

    tuples = {"X": tuple_x, "Y": tuple_y, "Z": tuple_z, "I": tuple_z}

    for qubit, operator in term:
        assert isinstance(qubit, int)
        if operator not in tuples:
            raise ValueError(f"Unknown operator {operator}")
        memory_map[alpha_label][qubit] = tuples[operator][0]
        memory_map[beta_label][qubit] = tuples[operator][1]
        memory_map[gamma_label][qubit] = tuples[operator][2]

    return memory_map


def pauli_term_to_preparation_memory_map(term: PauliTerm, label: str = "preparation") -> Dict[str, List[float]]:
    """
    Given a ``PauliTerm``, create a memory map corresponding to the ZXZXZ-decomposed single-qubit
    gates that prepare the plus one eigenstate of the ``PauliTerm``. For example, if we have the
    following program:

        RZ(preparation_alpha[0]) 0
        RX(pi/2) 0
        RZ(preparation_beta[0]) 0
        RX(-pi/2) 0
        RZ(preparation_gamma[0]) 0

    We can prepare the ``|+>`` state (by default we start in the ``|0>`` state) by providing the
    following memory map (which corresponds to ``RY(pi/2)``):

        {'preparation_alpha': [0.0], 'preparation_beta': [pi/2], 'preparation_gamma': [0.0]}

    :param term: The ``PauliTerm`` in question.
    :param label: The prefix to provide to ``pauli_term_to_euler_memory_map``, for labeling the
        declared memory regions. Defaults to "preparation".
    :return: Memory map for preparing the desired state.
    """
    return pauli_term_to_euler_memory_map(term, prefix=label, tuple_x=P_X, tuple_y=P_Y, tuple_z=P_Z)


def pauli_term_to_measurement_memory_map(term: PauliTerm, label: str = "measurement") -> Dict[str, List[float]]:
    """
    Given a ``PauliTerm``, create a memory map corresponding to the ZXZXZ-decomposed single-qubit
    gates that allow for measurement in the eigenbasis of the ``PauliTerm``. For example, if we
    have the following program:

        RZ(measurement_alpha[0]) 0
        RX(pi/2) 0
        RZ(measurement_beta[0]) 0
        RX(-pi/2) 0
        RZ(measurement_gamma[0]) 0
        MEASURE 0 ro[0]

    We can measure in the ``Y`` basis (by default we measure in the ``Z`` basis) by providing the
    following memory map (which corresponds to ``RX(pi/2)``):

        {'measurement_alpha': [pi/2], 'measurement_beta': [pi/2], 'measurement_gamma': [pi/2]}

    :param term: The ``PauliTerm`` in question.
    :param label: The prefix to provide to ``pauli_term_to_euler_memory_map``, for labeling the
        declared memory regions. Defaults to "measurement".
    :return: Memory map for measuring in the desired basis.
    """
    return pauli_term_to_euler_memory_map(term, prefix=label, tuple_x=M_X, tuple_y=M_Y, tuple_z=M_Z)


def merge_memory_map_lists(
    mml1: List[Dict[str, List[float]]], mml2: List[Dict[str, List[float]]]
) -> List[Dict[str, List[float]]]:
    """
    Given two lists of memory maps, produce the "cartesian product" of the memory maps:

        merge_memory_map_lists([{a: 1}, {a: 2}], [{b: 3, c: 4}, {b: 5, c: 6}])

        -> [{a: 1, b: 3, c: 4}, {a: 1, b: 5, c: 6}, {a: 2, b: 3, c: 4}, {a: 2, b: 5, c: 6}]

    :param mml1: The first memory map list.
    :param mml2: The second memory map list.
    :return: A list of the merged memory maps.
    """
    if not mml1:
        return mml2
    if not mml2:
        return mml1
    return [{**d1, **d2} for d1, d2 in itertools.product(mml1, mml2)]



from copy import copy
from dataclasses import dataclass
from typing import Union, Dict, List, Any, Optional, no_type_check

from pyquil.quilatom import (
    substitute,
    Expression,
    FormalArgument,
    Frame,
    MemoryReference,
    Parameter,
    Qubit,
    TemplateWaveform,
    WaveformReference,
)
from pyquil.quilbase import (
    AbstractInstruction,
    Capture,
    Declare,
    DelayFrames,
    DelayQubits,
    DefCalibration,
    DefMeasureCalibration,
    Fence,
    FenceAll,
    Gate,
    Measurement,
    Pragma,
    Pulse,
    RawCapture,
    ResetQubit,
    SetFrequency,
    SetPhase,
    SetScale,
    ShiftFrequency,
    ShiftPhase,
    SwapPhase,
)


class CalibrationError(Exception):
    pass


class CalibrationDoesntMatch(CalibrationError):
    pass


@dataclass
class CalibrationMatch:
    cal: Union[DefCalibration, DefMeasureCalibration]
    settings: Dict[Union[FormalArgument, Parameter], Any]


@no_type_check
def fill_placeholders(obj, placeholder_values: Dict[Union[FormalArgument, Parameter], Any]):
    """Update Parameter and FormalArgument references in objects with
    their corresponding definitions.

    It is an error if the object has a Parameter or FormalArgument reference without
    a corresponding definition in placeholder_values.

    :param obj: A Quil AST object.
    :param placeholder_values: A dictionary mapping placeholders to their values.
    :returns: The updated AST object.
    """
    try:
        if obj is None or isinstance(obj, (int, float, complex, Qubit, MemoryReference)):
            return obj
        elif isinstance(obj, Expression):
            # defer to the usual PyQuil substitution
            return substitute(obj, {k: v for k, v in placeholder_values.items() if isinstance(k, Parameter)})
        elif isinstance(obj, FormalArgument):
            return placeholder_values[obj]
        elif isinstance(obj, Frame):
            return Frame(fill_placeholders(obj.qubits, placeholder_values), obj.name)
        elif isinstance(obj, WaveformReference):
            return obj
        elif isinstance(obj, TemplateWaveform):
            return obj.__class__(**fill_placeholders(obj.__dict__, placeholder_values))
        elif isinstance(obj, list):
            return [fill_placeholders(elt, placeholder_values) for elt in obj]
        elif isinstance(obj, dict):
            return {k: fill_placeholders(v, placeholder_values) for (k, v) in obj.items()}
        elif isinstance(obj, tuple):
            return tuple([fill_placeholders(item, placeholder_values) for item in obj])
        elif isinstance(obj, Pragma) and obj.command == "LOAD-MEMORY":
            (source,) = obj.args
            arg = FormalArgument(obj.freeform_string)
            if arg in placeholder_values:
                return Pragma("LOAD-MEMORY", [source], str(placeholder_values[arg]))
            else:
                return obj
        else:
            specs = {
                Gate: ["params", "qubits"],
                Measurement: ["qubit", "classical_reg"],
                ResetQubit: ["qubit"],
                Pulse: ["frame", "waveform"],
                SetFrequency: ["frame", "freq"],
                ShiftFrequency: ["frame", "freq"],
                SetPhase: ["frame", "phase"],
                ShiftPhase: ["frame", "phase"],
                SwapPhase: ["frameA", "frameB"],
                SetScale: ["frame", "scale"],
                Capture: ["frame", "kernel", "memory_region"],
                RawCapture: ["frame", "duration", "memory_region"],
                DelayQubits: ["qubits", "duration"],
                DelayFrames: ["frames", "duration"],
                Fence: ["qubits"],
                FenceAll: [],
                Declare: [],
                Pragma: [],
            }
            if type(obj) in specs:
                attrs = specs[type(obj)]
                updated = copy(obj)
                for attr in attrs:
                    setattr(updated, attr, fill_placeholders(getattr(updated, attr), placeholder_values))
                return updated
            else:
                raise CalibrationError(f"Unable to fill placeholders in object {obj}.")
    except Exception as e:
        raise e
        # raise ValueError(f"Unable to fill placeholders in  object {obj}.")


def match_calibration(
    instr: AbstractInstruction, cal: Union[DefCalibration, DefMeasureCalibration]
) -> Optional[CalibrationMatch]:
    """Match a calibration definition to an instruction.

    On a successful match, return a (possibly empty) dictionary mapping calibration
    arguments and parameters to their values.

    On a failure, return None.
    """
    settings: Dict[Any, Any] = {}

    @no_type_check
    def unpack_field(cal_field, instr_field):
        if isinstance(cal_field, (Parameter, FormalArgument)):
            if instr_field is None:
                raise CalibrationDoesntMatch()
            settings[cal_field] = instr_field
        elif cal_field != instr_field:
            raise CalibrationDoesntMatch()

    if isinstance(instr, Measurement) and isinstance(cal, DefMeasureCalibration):
        try:
            unpack_field(cal.qubit, instr.qubit)
            unpack_field(cal.memory_reference, instr.classical_reg)
            return CalibrationMatch(cal, settings)
        except CalibrationDoesntMatch:
            return None
    elif isinstance(instr, Gate) and isinstance(cal, DefCalibration):
        try:
            unpack_field(cal.name, instr.name)
            for cal_param, instr_param in zip(cal.parameters, instr.params):
                unpack_field(cal_param, instr_param)
            for cal_arg, instr_arg in zip(cal.qubits, instr.qubits):
                unpack_field(cal_arg, instr_arg)
            return CalibrationMatch(cal, settings)
        except CalibrationDoesntMatch:
            return None
    else:
        return None


def expand_calibration(match: CalibrationMatch) -> List[AbstractInstruction]:
    """" Expand the body of a calibration from a match. """
    return [fill_placeholders(instr, match.settings) for instr in match.cal.instrs]



##############################################################################
# Copyright 2016-2019 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

import warnings
from typing import Optional

from pyquil.latex._diagram import DiagramSettings
from pyquil.quil import Program


def to_latex(circuit: Program, settings: Optional[DiagramSettings] = None) -> str:
    from pyquil.latex._main import to_latex

    warnings.warn(
        '"pyquil.latex.latex_generation.to_latex" has been moved -- please import it'
        'as "from pyquil.latex import to_latex going forward"',
        FutureWarning,
    )
    return to_latex(circuit, settings)
